#!/usr/bin/perl

$| = 1;

# ddrsummarize.pl   ("summarize a ddrescue log file")
#
# This is a Perl program which reads a GNU ddrescue 1.x "log files" (really,
# sector list files), and prints a summary of the number of sectors in each
# catagory.


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 26-Oct-2005 - first version.  -DAB

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "9, 05-Oct-09";
# the TLIB Version Control version number and last-modified date


if (!defined $verbose) {
   $verbose = 0;
}

$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


# Handle input parameter and open the file:

while (($#ARGV >= 0) && ($ARGV[0] =~ /^-v/)) {
   shift @ARGV;
   $verbose++;
}
($log_input) = @ARGV;

if ((0 > $#ARGV) || !defined $log_input) {
   die "ddrsummarize.pl  (\"summarize a ddrescue log file\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage:\n" .
       "   perl -w ddrsummarize.pl inputfile\n" .
       "\n" .
       "or, for verbose output:\n" .
       "   perl -w ddrsummarize.pl -v inputfile\n";
     #  "\n" .
     #  "'-v' can be repeated for very verbose output.\n";
}

if ((!defined $loaded_via_REQUIRE) && $verbose) {
   print "You are using Perl version $perlver\n";
   ($sec,$min,$hr,$day,$mon,$year,$wday, $yday,$isdst) = localtime(time);
   printf "%4d-%02d-%02d %02d:%02d:%02d\n", $year+1900, $mon+1, $day, $hr, $min, $sec;
}


######################################################################
# Here's a sample of what a ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.1
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby_512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}


# convert a really big integer into a string with commas
sub commas {
   local( $i ) = @_;
   local( $result, $x );
   if ($i <= 9999.5) {
      $result = sprintf( "%4.0f", $i );
   } else {
       $x = sprintf( "%1.0f", $i );
       $result = '';
       $x = reverse( $x );
       while ('' ne $x) {
          if ($result ne '') {
             $result .= ',';
          }
          $result .= substr( $x, 0, 3 );
          if (length($x) > 3) {
             $x = substr( $x, 3 );
          } else {
             $x = '';
          }
      }
      $result = reverse( $result );
   }
   # printf "dbg: %1.0f = %s\n", $i, $result;
   return $result;
}


sub process1file {
   local( $log_input ) = @_;
   local( $numranges, $total_num_sectors, %counts, @tmp, %tmp2, $pos_byt, $pos_sect, $len_byt, $status, $len_sect, $bytes, $megabytes, $percentage, $unrecovered, $unrecovered2 );
   if (!open( LOG_INPUT, "<$log_input")) {
      print STDERR "ERROR: Could not read $log_input, $!\n";
   } else {
      print "$log_input :\n";

      # read log_input
      $numranges = 0;
      $total_num_sectors = 0;
      $previous_status = ' ';
      while (<LOG_INPUT>) {
         if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
            $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
            # Parse a ddrescue log file line:
            @tmp = split( /\s+/, $_ );
            # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
            ($pos_byt, $len_byt, $status) = @tmp;
            if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
               if ((1 == $#tmp) && (0 == $numranges) && ($len_byt =~ /^[\?\*\/\-\+]$/)) {
                  # the new status line (introduced in ddrescue 1.6 or 1.7)
                  $pos_sect = &divby_512($pos_byt);
                  $status = $len_byt;
                  print "Status: ";
                  if ($status eq '?') {
                     print "copying data";
                  } elsif ($status eq '*') {
                     print "trimming bad areas";
                  } elsif ($status eq '/') {
                     print "splitting bad areas";
                  } elsif ($status eq '-') {
                     print "retrying bad blocks";
                  } elsif ($status eq '+') {
                     print "finished";
                  } else {
                     print "unrecognized status '$status'";
                  }
                  print " (status code = '$status'), at sector $pos_sect\n";
               } else {
                  print "ERROR: unrecognized line in '$log_input':\n$_\n";
                  printf 'dbg: $#tmp=%d, $numranges=%d, $len_byt="%s"' . "\n", $#tmp, $numranges, $len_byt;
               }
               next;
            }
            if ($len_byt eq '-1') {
               print "Note: length=-1, line ignored:  $_\n";
               next;
            }
            # $pos_sect = &divby_512($pos_byt);
            # convert byte count to sector count:
            $len_sect = &divby_512($len_byt);

            if (!defined $counts{$status}) {
               $counts{$status} = 0;
               $numintervals{$status} = 0;
            }
            $counts{$status} += $len_sect;
            $total_num_sectors += $len_sect;
            $numranges++;
            if ($status ne $previous_status) {
               $numintervals{$status}++;
               $previous_status = $status;
            }
         }
      }
      close LOG_INPUT;

      %tmp2 = %counts;
      foreach $status ('+','-','/',keys %counts) {
         if (defined $tmp2{$status}) {
            printf "$status%11d sectors =", $counts{$status};
            $bytes = &commas( $counts{$status} * 512.0 );
            $megabytes = $counts{$status} / 2048;
            $percentage = 100.0 * ($counts{$status} / $total_num_sectors);
            printf "%17s bytes =%10.2f MB = %6.3f%% of total", $bytes, $megabytes, $percentage;
            if ($verbose) {
               printf ", in %4d intervals", $numintervals{$status};
            }
            print "\n";
            delete $tmp2{$status};
         }
      }
      printf "Total:\n" .
             "  %10d sectors =%17s bytes =%10.2f MB (in %d intervals)\n",
             $total_num_sectors, &commas($total_num_sectors*512.0), $total_num_sectors/2048, $numranges;
      $unrecovered2 = 0;
      if (defined $counts{'-'}) {
         $unrecovered2 += $counts{'-'};  # definitely bad
      }
      if (defined $counts{'/'}) {
         $unrecovered2 += $counts{'/'};  # unsplit bad area
      }
      if (defined $counts{'?'}) {
         $unrecovered2 += $counts{'?'};  # untried
      }
      if (defined $counts{'*'}) {
         $unrecovered2 += $counts{'*'};  # untrimmed bad area
      }
      $unrecovered = $total_num_sectors;
      if (defined $counts{'+'}) {
         $unrecovered -= $counts{'+'};
      }
      printf "Unrecovered:";
      if ($unrecovered2 != $unrecovered) {
         print "           (unrecovered sector count is $unrecovered, or maybe $unrecovered2)";
      }
      printf "\n" .
             "%12d sectors =%17s bytes =%10.2f MB = %6.3f%% of total\n",
             $unrecovered, &commas($unrecovered*512.0), $unrecovered/2048,
             100.0*($unrecovered/$total_num_sectors);
      print "\n"; #extra blank line at end
   }
}

foreach $inpfile (@ARGV) {
   &process1file( $inpfile );
}

1; #return true, in case invoked via 'require' from ddrcombine.pl

__END__

