#!/usr/bin/perl -w

# FAT32INF.PL -- a program to display FAT32 file system information.
# (Written in Perl 4.036, and compatible with Perl 5.)
#
#
# Copyright 2007, by David A. Burton.
# Burton Systems Software
# http://www.burtonsys.com/
# POB 4157, Cary, NC 27519-4157
# Tel: +1 919-481-0149
# Email: dave at burtonsys.com
#
#

$ver_ordinal_and_date = $b_bootstrap = $b_unused1 = $b_unused2 = $b_unused3 = 0;  # shut up Perl warnings

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "5, 20-Jan-07";
# the TLIB Version Control version number and last-modified date


# Here are a couple of descriptions of the FAT32 file system format:
#   http://www.maverick-os.dk/FileSystemFormats/FAT32_FileSystem.html
#   http://home.att.net/~rayknights/pc_boot/w95bboot.htm
#   http://home.no.net/tkos/info/fat32.html


$| = 1;   # predefined variable. If <> 0 then each print to the console
          # will immediatly be displayed, instead of buffered.

$copr = "Copyright 2007, David A. Burton, Burton Systems Software.   www.burtonsys.com\n";
print "fat32inf.pl -- FAT32 File Sector Information Utility.   v.$ver_ordinal_and_date\n" . $copr;



$o_verbose = 0;   # '-v' sets it to 1, -vv or -v2 or "-v -v" sets it to 2
$o_debugmode = 0;   # '-D' sets it to 1, -DD or -D2 or "-D -D" sets it to 2
$o_minmem = 0;   # '-m' sets it to 1
$o_showbootsect = 0;  # '-b' sets it to 1
$o_showFAT = 0;  # '-f' sets it to 100, '-F' sets it to 1000000
$o_listfiles = 0;  # '-l' sets it to 1
$o_rawlistfiles = 0; # '-L' sets it to 1
$o_search4bootsect = 0;  # '-h' sets it to 1
$o_lookup_sector = 0;  # '-s' sets it to 1
$o_listdeletedfiles = 0;  # '-d' sets it to 1

# initialize globals:
$lba_of_bootsect = -1;  # which boot sector is currently loaded


# Perl pack/unpack format representing the structure of the FAT32 Boot Sector:
$BootSector_fmt =
  'a3'  . # 000h  3 bytes  JMP 0x80h
  'a8'  . # 003h  8 bytes  Oem ID - Name of the formatting OS
          ## BIOSParameter Block (00Bh - 03Fh):
  'S'   . # 00Bh  2 bytes  Bytes per Sector on the physical medium - 512/1024/2048/4096, but normally 512 bytes
  'C'   . # 00Dh  1 bytes  Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors
  'S'   . # 00Eh  2 bytes  Reserved sectors in front of the FAT(s) incl. the Boot sector (usually 32 for FAT32, 1 for FAT12FAT16, 0 for NTFS)
  'C'   . # 010h  1 bytes  Number of FAT copies - Normally 2 for FAT, 0 for NTFS
  'a4'  . # 011h  4 bytes  Not used in FAT32 or NTFS (1st 2 bytes are number of root directory entries in FAT12/FAT16, 2nd 2 bytes can be a 16-bit total number of data sectors for FAT12/FAT16, if that value fits in 16 bits)
  'C'   . # 015h  1 bytes  Media Descriptor - The same as in FAT16, but FAT32 is only allowed on harddrives, so the value is F8h
  'S'   . # 016h  2 bytes  Not used in FAT32 or NTFS (number of sectors/fat in FAT12/FAT16)
  'S'   . # 018h  2 bytes  Sectors per Track - The disc geometry used when formatting the partition
  'S'   . # 01Ah  2 bytes  Heads - The disc geometry used when formatting the partition
  'L'   . # 01Ch  4 bytes  Supposedly the number of "hidden sectors," but actually usually the sector number of the partition boot sector relative to the partition table sector that points to the partition.  Usually 63 for extended partitions, and for the first primary partition; usually larger for other primary partitions.
  'L'   . # 020h  4 bytes  Number of sectors in the partition (for FAT32 and some FAT16 partitions)
  # Note: offsets 24-64 are defined for FAT32 only:
  'L'   . # 024h  4 bytes  Number of sectors per FAT (FAT32 only)
  'S'   . # 028h  2 bytes  FAT32 handling flags
  'S'   . # 02Ah  2 bytes  FAT32 Drive Version (High byte = Major version, Low byte = Minor version)
  'L'   . # 02Ch  4 bytes  Cluster number for the start of the FAT32 Root Directory Table (usually 2)
  'S'   . # 030h  2 bytes  Sector number of File System Information Sector, BIGFATBOOTFSINFO struct
  'S'   . # 032h  2 bytes  Sector number of Backup Boot Sector
  'a12' . # 034h 12 bytes  Reserved
          ## Extended BIOS Parameter Block (040h - 059h):
  'C'   . # 040h  1 bytes  Logical Drive Number - Normally 00h for floppies and 80h for hard drives.
  'C'   . # 041h  1 bytes  Current Head
  'C'   . # 042h  1 bytes  Signature byte, should be 29h to indicate that Serial Number, Volume Label, and System ID are valid
  'L'   . # 043h  4 bytes  Serial Number (randomly generated)
  'a11' . # 047h 11 bytes  Volume Label - The same as stored in a special file in the root directory.
  'a8'  . # 052h  8 bytes  System ID - This is the string 'FAT32 '
          ## Code:
  'a420'. # 05Ah 420 bytes possibly bootstrap loader codeb_unused2
  'S';    # 1FEh  2 bytes  Executable sector signature (AA55h)


# Unpack a boot sector using $BootSector_fmt as the template:
sub unpack_a_boot_sector {
  local( $onesector ) = @_;
  local( @results ) = unpack( $BootSector_fmt, $onesector );
  return @results;
}


# @sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

sub do1option {
  local($ch,$valu) = @_;
  if ('v' eq $ch) {
     $o_verbose+= $valu;
  } elsif ('D' eq $ch) {
     $o_debugmode+= $valu;
  } elsif ('m' eq $ch) {
     $o_minmem+= $valu;
  } elsif ('b' eq $ch) {
     $o_showbootsect+= $valu;
  } elsif ('f' eq $ch) {
     $o_showFAT += 100*$valu;
  } elsif ('F' eq $ch) {
     $o_showFAT += 1000000;
  } elsif ('l' eq $ch) {
     $o_listfiles += $valu;
  } elsif ('L' eq $ch) {
     $o_rawlistfiles += $valu;
  } elsif ('d' eq $ch) {
     $o_listdeletedfiles += $valu;
  } elsif ('h' eq $ch) {
     $o_search4bootsect+= $valu;
  } elsif ('s' eq $ch) {
     $o_lookup_sector+= $valu;
  } elsif ('-' ne $ch) {
     print "Warning: unrecognized option (\"-$ch\") ignored.\n";
  }
  # Note: if you add another option, be sure to fix the 'while' condition, too
}


# Check for command-line options.
# These are equivalent (where x, y, and z are any legal options):
#    -x -y -z
#    -xyz
#    -x1 -y1 -z1
# These are also equivalent:
#    -xx
#    -x2
# Etc.
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[vdDmbfFlLs]/)) {
   if ($ARGV[0] =~ /^-.\d/) {
      $ch = substr( $ARGV[0], 1, 1 );
      $valu = 0 + substr( $ARGV[0], 2 );
      # print "dbg: ch='$ch', valu='$valu'\n";
      &do1option( $ch, $valu );
   } else {
      foreach $ch (split(//,$ARGV[0])) {
         &do1option( $ch, 1 );
      }#foreach
   }
   shift @ARGV;
}


$numargs = $#ARGV + 1;
if ($o_search4bootsect) {
   if ($numargs != 3) {
      &givehelp;
   }
} elsif ($o_lookup_sector) {
   if ($numargs < 3) {
      &givehelp;
   }
} elsif ($numargs != 2) {
   &givehelp;
}
if ($o_search4bootsect && ($o_lookup_sector || $o_listfiles || $o_rawlistfiles)) {
   print "ERROR: You may not specify -h with -l, -L, or -s\n";
   &givehelp;
}

sub givehelp {
   print 'ERROR: $#ARGV = ' . $#ARGV . "\n\n";
   print "Usage (basic info mode):\n" .
         "  perl fat32inf.pl {options} drive bootsect\n" .
         "or (search for boot sectors):\n" .
         "  perl fat32inf.pl -h drive firstsect lastsect\n" .
         "or (look up files from sector numbers):\n" .
         "  perl fat32inf.pl -s drive bootsect sect1 sect2 sect3...\n" .
         "\n" .
         "Where:\n" .
         "  'drive' is a device name or a file containing a drive image, and\n" .
         "  'bootsect' is the LBA sector number (0-based) of the partition boot sector\n" .
         "  'firstsect' and 'lastsect' specify a range of sector numbers; those which\n" .
         "           resemble boot sectors will be displayed\n" .
         "  'sect1', 'sect2', etc. are LBA sector numbers to report the containing files for.\n" .
         "  {options} are optional:\n" .
         "    -m    reduced Memory usage (usually makes little difference)\n" .
         "    -s    lookup filenames from LBA Sector numbers (prefix with '0x' for hexidecimal)\n" .
         "    -h    Heuristic search for boot sectors\n" .
         "    -b    display Boot sector in human-understandable format\n" .
         "    -f    display the first 100 sectors of the main FAT (-ff for 200, etc.)\n" .
         "    -F    display the entire main FAT (probably very long!)\n" .
         "    -l    List all files, regular format (note: '-l' is lower-case '-L', not '-1')\n" .
         "    -L    List all files, raw directory format\n" .
         "    -LL (or -L2)  List all files + long name fragments, raw directory format\n" .
         "    -d    List deleted files, raw directory format\n" .
         "    -dd (or -d2)  List deleted files + long name fragments, raw directory format\n" .
         "    -v    Verbose mode\n" .
         "    -D    Debug mode prints (or '-dd' or '-d2' for even more debug prints)\n" .
         "\n" .
         "(Note: often the main boot sector is sector 63, and the backup boot\n" .
         "sector is sector 69.)\n" .
         "\n" .
         "Examples:\n" .
         "  perl fat32inf.pl -b /dev/hdb1 0           (display partition boot sector)\n" .
         "  perl fat32inf.pl -bl /dev/hdb1 0          (display boot sector and list files)\n" .
         "  perl fat32inf.pl -b -l /dev/hdb1 0        (same as previous example)\n" .
         "  perl fat32inf.pl -s /dev/hdb1 0 0x19E6    (look up sector #6330, show file name)\n" .
         "  perl fat32inf.pl -s /dev/hdb 63 6630      (same, if hdb1 is at LBA 63)\n" .
         "  perl fat32inf.pl -s drive.ima 63 0x18E6   (same, if drive.ima is copy of /dev/hdb)\n" .
         "  perl fat32inf.pl -s drive.ima 63 6630 19276 22040  (look up 3 sectors at once)\n" .
         "\n" .
         $copr;
   exit 1;
}


sub hex_or_positive_decimal {
  local( $str ) = @_;
  local( $result ) = 0;
  if ($str =~ /^[0-9]+$/) {
     $result = 0+$str;
  } elsif ($str =~ /^0x[0-9a-fA-F]+$/) {
     $result = eval $str;
  } else {
     print "ERROR: '$str' is not a positive integer or C-format hexidecimal integer.\n";
     exit 2;
  }
  if ($o_debugmode) {
     print "dbg: '$str' = $result\n";
  }
  return $result;
}# sub hex_or_positive_decimal



# some command-line parameter processing:

# first two parameters are common to all modes:
$drive = $ARGV[0];
$lba_offset1 = $ARGV[1];
$lba_offset1 = &hex_or_positive_decimal( $lba_offset1 );
# handle 3rd parameter for '-h' mode
if ($o_search4bootsect) {
   # '-h' (heuristic search for boot sectors)
   $lba_offset2 = $ARGV[2];
   $lba_offset2 = &hex_or_positive_decimal( $lba_offset2 );
   if ($lba_offset2 < $lba_offset1) {
      print "ERROR: 2rd parameter (\"$lba_offset1\") must be <= 3nd parameter (\"$lba_offset2).\n";
      exit 2;
   }
} else {
   undef $lba_offset2;
}


# unpack the interesting fields of a boot sector into global variables
sub unpack_boot_sect_to_globals {
  local( $fbuf ) = @_;
  local(@bootsectorfields) = &unpack_a_boot_sector( $fbuf );
  (  $b_jmp80                 , # 000h  3 bytes  JMP 0x80h
     $b_OemId                 , # 003h  8 bytes  Oem ID - Name of the formatting OS
                                ## BPB: BIOS Parameter Block (00Bh - 03Fh):
     $b_bytesPerSector        , # 00Bh  2 bytes  Bytes per Sector on the physical medium - Normally 512 bytes
     $b_sectorsPerCluster     , # 00Dh  1 bytes  Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors
     $b_resSectBeforeFAT      , # 00Eh  2 bytes  Reserved sectors in front of the FAT(s) incl. the Boot sector (usually 32 for FAT32, 0 for NTFS)
     $b_numFATs               , # 010h  1 bytes  Number of FAT copies - Normally 2
     $b_unused1               , # 011h  4 bytes  Not used in FAT32 or NTFS
     $b_F8h                   , # 015h  1 bytes  Media Descriptor - The same as in FAT16, but FAT32 is only allowed on harddrives, so the value is F8h
     $b_unused2               , # 016h  2 bytes  Not used in FAT32 or NTFS
     $b_sectorsPerTrack       , # 018h  2 bytes  Sectors per Track - The disc geometry used when formatting the partition
     $b_heads                 , # 01Ah  2 bytes  Heads - The disc geometry used when formatting the partition
     $b_numHiddenSectors      , # 01Ch  4 bytes  supposedly number of sectors on the disk from start of the partition to beginning of the first FAT
     $b_sectorsInPartition    , # 020h  4 bytes  Number of sectors in the partition
                                ## FAT32-only BPB fields:
     $b_sectorsPerFAT         , # 024h  4 bytes  Sectors per FAT
     $b_handlingFlags         , # 028h  2 bytes  FAT handling flags
     $b_driverVersion         , # 02Ah  2 bytes  FAT32 Driver Version (High byte = Major version, Low byte = Minor version)
     $b_clusterOfRootDirStart , # 02Ch  4 bytes  Cluster number for the start of the Root Directory
     $b_sectorOfFileSystInf   , # 030h  2 bytes  Sector number of File System Information Sector, BIGFATBOOTFSINFO struct
     $b_sectorOfBackupBootSect, # 032h  2 bytes  Sector number of Backup Boot Sector (often 6)
     $b_unused3               , # 034h 12 bytes  Reserved
                                ## Extended BIOS Parameter Block (040h - 059h):
     $b_logicalDriveNum       , # 040h  1 bytes  Logical Drive Number - Normally 00h for floppies and 80h for hard drives.
     $b_currentHead           , # 041h  1 bytes  Current Head
     $b_signature1            , # 042h  1 bytes  Signature (should be 29h)
     $b_serialNumber          , # 043h  4 bytes  Randomly generated serial number
     $b_volumeLabel           , # 047h 11 bytes  Volume Label - The same as stored in a special file in the root directory.
     $b_systemID              , # 052h  8 bytes  System ID - This is the string 'FAT32 '
                                ## Code:
     $b_bootstrap             , # 05Ah 420 bytes possibly bootstrap loader code
     $b_AA55h                   # 1FEh  2 bytes  Executable sector signature (AA55h)
  ) = @bootsectorfields;
} # sub unpack_boot_sect_to_globals


# Calculate various relative sector numbers, which will come in handy later.
# We assume that &unpack_boot_sect_to_globals() has already been called, and
# that $lba_of_bootsect has already been set.
sub calc_offsets {
  # Calculate the sector number of the first sector of the first data cluster
  # (which is cluster number 2, oddly enough), relative to the boot sector:
  $rel_first_data_sector = $b_resSectBeforeFAT + ($b_numFATs * $b_sectorsPerFAT);
  # (Note: for FAT12 & FAT16 you'd have to also add the root directory size,
  # because the FAT12/16 root directory is not stored in the data clustors.)

  # Calculate the data-area-base sector number, relative to the boot sector.
  # It would be the starting sector number of cluster 0, if there were such a
  # thing as cluster 0.
  $rel_sect_of_cluster0 = $rel_first_data_sector - (2 * $b_sectorsPerCluster);
  if ($o_debugmode) {
     print "dbg: b_resSectBeforeFAT=$b_resSectBeforeFAT,  b_numFATs=$b_numFATs, b_sectorsPerFAT=$b_sectorsPerFAT.\n";
     print "     rel_first_data_sector = b_resSectBeforeFAT + (b_numFATs * b_sectorsPerFAT) = $b_resSectBeforeFAT+($b_numFATs*$b_sectorsPerFAT) = $rel_first_data_sector.\n";
     print "dbg: rel_sect_of_cluster0 = rel_first_data_sector-(2*b_sectorsPerCluster) = $rel_first_data_sector-(2*$b_sectorsPerCluster) = $rel_sect_of_cluster0.\n";
  }

  # Calculate the locations of the two FATs:
  $rel_sect_of_FAT1 = $b_resSectBeforeFAT;
  if ($b_numFATs >= 2) {
     $rel_sect_of_FAT2 = $rel_sect_of_FAT1 + $b_sectorsPerFAT;
  } else {
     $rel_sect_of_FAT2 = $rel_sect_of_FAT1;
  }

  # Calculate number of data clusters:
  if ($b_sectorsPerCluster <= 0) {
     $num_data_clusters = $highest_cluster_num = 0;
  } else {
     $num_data_clusters = int( ($b_sectorsInPartition - $rel_first_data_sector) / $b_sectorsPerCluster );
     $highest_cluster_num = $num_data_clusters + 1;  # $num_data_clusters numbered 2 .. $highest_cluster_num
  }

  local( $lastdatasect ) = (&cluster_to_lba($highest_cluster_num));
  if ($highest_cluster_num >= (128 * $b_sectorsPerFAT)) {
     if ($o_debugmode) {
        print "ERROR: highest cluster number is too large for the FAT size.\n";
        print "Adjusting it down from $highest_cluster_num to";
     }
     $highest_cluster_num = int($b_sectorsPerFAT / 128);
     if ($o_debugmode) {
        print " $highest_cluster_num\n";
     }
  }
  if ($o_debugmode) {
     print "highest_cluster_num=$highest_cluster_num, lastdatasect= LBA $lastdatasect, b_sectorsInPartition=$b_sectorsInPartition\n";
  }
} # sub calc_offsets


# Test a sector to see how much it resembles a partition boot sector.
# Input is the LBA sector number of possible boot sector.
# Output is a score, 0..10, with 10 being definitely a boot sector and
# zero being not at all resembling a boot sector.
sub test4bootsect {
  local( $lba_offset ) = @_;
  local( $score ) = 0;  # result: an indication of the liklihood that it is a boot sector
  &ensure_bootsect_loaded( $lba_offset );
  if ($b_AA55h == 0x0AA55) {
     $score += 2.7;
  }
  if ($b_jmp80 =~ /^(\xEB|\xE9)/) {
     # jmp instruction can begin with 0EBh or 0E9h
     $score += 0.5;
  }
  if (($b_numFATs >= 0) || ($b_numFATs <= 2)) {
     $score += 0.3;
  }
  if ($b_F8h == 0xF8) {
     $score += 0.8;
  }
  if ($b_bytesPerSector == 512) {
     # usual case
     $score += 1.4;
  } elsif (($b_bytesPerSector == 1024) || ($b_bytesPerSector == 2048) || ($b_bytesPerSector == 4096)) {
     # legal but rare
     $score += 1.3;
  }
  if (($b_sectorsPerCluster == 1) || ($b_sectorsPerCluster == 2) || ($b_sectorsPerCluster == 4) ||
      ($b_sectorsPerCluster == 8) || ($b_sectorsPerCluster == 16) || ($b_sectorsPerCluster == 32) ||
      ($b_sectorsPerCluster == 64) || ($b_sectorsPerCluster == 128)) {
     $score++;
  }
  if (($b_sectorOfBackupBootSect == 6) || ($b_sectorOfBackupBootSect == 0)){
     # seems to usually be 6 for FAT32 but 0 for NTFS
     $score += 0.3;
  }
  if ($b_OemId =~ /^NTFS/) {
     $score += 2.5;
     if ($b_resSectBeforeFAT == 0) {
        $score += 0.5;
     }
  } else {
     if ($b_OemId =~ /^(MSWIN|IBM|MSDOS)/) {
        $score += 0.5;
     }
     if ($b_systemID =~ /^FAT32/) {
        $score++;
     }
     if ($b_signature1 == 0x29) {
        $score++;
     }
     if ($b_resSectBeforeFAT == 1) {
        # usual value for FAT12/16 is 1
        $score += 0.1;
     } elsif ($b_resSectBeforeFAT == 32) {
        # usual value for FAT32 is 32
        $score += 0.5;
     }
  }
  if ($highest_cluster_num < 65526) {
     $score -= 1.0;  # too small to be a legal FAT32 volume.
     if ($o_debugmode) {
        print "This is too small to be a legal FAT32 volume.  Highest cluster number is only $highest_cluster_num, but it must be >= 65526.\n";
     }
  }
  return $score;
} #sub test4bootsect


# Display FAT32 file system info gleaned from a partition boot sector.
# Inputs are:
#   $fbuf             - a 512-byte sector, which might be a boot sector
#   $lba_offset       - the LBA disk address from which $fbuf was read
sub display_bootsect {
  local( $lba_offset ) = @_;
  local( $fileofs ) = (512 * $lba_offset);
  local( $hexfileofs ) = sprintf( "%X", $fileofs );
  if ('FFFFFFFF' eq $hexfileofs) {
      $hexfileofs = "";  # too big, can't display it
  } else {
      $hexfileofs = "= 0x0" . $hexfileofs;
  }
  printf "LBA $lba_offset = 0x0%X  -->  File offset of boot sector is: $fileofs $hexfileofs\n", $lba_offset;
  &ensure_bootsect_loaded( $lba_offset );
  printf "000h  3 bytes  Jmp                    = %02X %02X %02X\n", ord(substr($b_jmp80,0,1)), ord(substr($b_jmp80,1,1)), ord(substr($b_jmp80,2,1));
  printf "003h  8 bytes  OemId                  = '%-8s' (Oem ID - Name of formatting OS)\n", $b_OemId;
  printf "--- BIOSParameter Block (00Bh - 03Fh):\n";
  printf "00Bh  2 bytes  bytesPerSector         = %-6d     (Bytes per Sector on physical medium - Normally 512 bytes)\n", $b_bytesPerSector;
  printf "00Dh  1 bytes  sectorsPerCluster      = %-4d       (Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors)\n", $b_sectorsPerCluster;
  printf "00Eh  2 bytes  resSectBeforeFAT       = %-6d     (Reserved sectors in front of FAT(s) incl. Boot sector)\n", $b_resSectBeforeFAT;
  printf "010h  1 bytes  numFATs                = %-4d       (Number of FAT copies - Normally 2 for FAT, 0 for NTFS)\n", $b_numFATs;
  printf "015h  1 bytes  F8h                    = 0x%02X       (Media Descriptor - should be F8h)\n", $b_F8h;
  printf "018h  2 bytes  sectorsPerTrack        = %-6d     (Sectors per Track - disc geometry used when formatting partition)\n", $b_sectorsPerTrack;
  printf "01Ah  2 bytes  heads                  = %-6d     (Heads - disc geometry used when formatting partition)\n", $b_heads;
  printf "01Ch  4 bytes  numHiddenSectors       = %-10d (sector no. of partition boot sector relative to the partition table sector that points to it)\n", $b_numHiddenSectors;
  printf "020h  4 bytes  sectorsInPartition     = %-10d (Number of sectors in partition)\n", $b_sectorsInPartition;
  printf "024h  4 bytes  sectorsPerFAT          = %-8d   (Sectors per FAT)\n", $b_sectorsPerFAT;
  printf "028h  2 bytes  handlingFlags          = 0x%04X     (FAT handling flags)\n", $b_handlingFlags;
  printf "02Ah  2 bytes  driverVersion          = 0x%04X     (FAT32 Driver Version (High byte = Major version, Low byte = Minor version))\n", $b_driverVersion;
  printf "02Ch  4 bytes  clusterOfRootDirStart  = %-6d     (Cluster number for start of Root Directory, usually 2)\n", $b_clusterOfRootDirStart;
  printf "030h  2 bytes  sectorOfFileSystInf    = %-4d       (Sector number of File System Information Sector, BIGFATBOOTFSINFO struct)\n", $b_sectorOfFileSystInf;
  printf "032h  2 bytes  sectorOfBackupBootSect = %-4d       (Sector number of Backup Boot Sector)\n", $b_sectorOfBackupBootSect;
  printf "--- Extended BIOS Parameter Block (040h - 059h):\n";
  printf "040h  1 bytes  logicalDriveNum        = 0x%02X       (Logical Drive Number - Normally 00h for floppies and 80h for hard drives.)\n", $b_logicalDriveNum;
  printf "041h  1 bytes  currentHead            = %-4d       (Current Head)\n", $b_currentHead;
  printf "042h  1 bytes  signature1             = 0x%02X       (Signature - should be 29h)\n", $b_signature1;
  printf "043h  4 bytes  serialNumber           = 0x%08X (ID - Random generated serial number)\n", $b_serialNumber;
  printf "047h 11 bytes  volumeLabel            = '%-11s' (Volume Label - same as stored in a special file in root directory.)\n", $b_volumeLabel;
  printf "052h  8 bytes  systemID               = '%-8s' (System ID - Should be the string 'FAT32 ')\n", $b_systemID;
  printf "--- Signature:\n";

  printf "1FEh  2 bytes  AA55h                  = 0x%04X     (Executable sector signature, should be 0xAA55h)\n", $b_AA55h;
} #sub display_bootsect


# Read a sector from specified LBA sector number
sub read_lba_sect {
  local( $lba_offset ) = @_;
  local( $fileofs ) = (512 * $lba_offset);
  local( $fbuf, $rc );
# print "dbg0/3: seek(,$fileofs,0); sysread(,,512)\n";
  if (!seek( DISKIMAGE, $fileofs, 0 )) {
     print "ERROR: Seek(,$fileofs,0) failed, $!\n";
     exit 2;
  }
  $rc = sysread(DISKIMAGE, $fbuf, 512);
  if (512 != $rc) {
     print "ERROR: Could not read '$drive' at ofs=$fileofs, $!\n";
     exit 2;
  }
  return $fbuf;
} # sub read_lba_sect


# Call this to make sure that the main boot sector for the partition of
# interest has been read, that global $lba_of_bootsect has been set, and
# that &unpack_boot_sect_to_globals() and &calc_offsets have been called
# to set various global variables from the boot sector contents.
sub ensure_bootsect_loaded {
  local( $lba_offset ) = @_;
  if ($lba_offset != $lba_of_bootsect) {
     # print "dbg: ensure_bootsect_loaded: read LBA $lba_offset.\n";
     $bootsect = &read_lba_sect( $lba_offset );  # bootsect is a global side-effect
     $lba_of_bootsect = $lba_offset;  # global side-effect (prevents unnecessary rereading)
     &unpack_boot_sect_to_globals( $bootsect );  # lots more global side-effects
     &calc_offsets;  # even more global side-effects
  }
} # sub ensure_bootsect_loaded


# Read a FAT32 boot sector, compare it to its backup, score each with various
# sanity checks, and report on the results.
sub check_bootsect_against_backup {
  local( $lba_offset ) = @_;
  &ensure_bootsect_loaded( $lba_offset );
  local( $fbuf1 ) = $bootsect;
  local( $score1 ) = &test4bootsect( $lba_offset );
  $b_sectorOfBackupBootSect1 = $b_sectorOfBackupBootSect;  # normally it is 6
  local( $lba_of_backup ) = 6 + $lba_offset;
  &ensure_bootsect_loaded( $lba_of_backup );
  local( $fbuf2 ) = $bootsect;
  local( $score2 ) = &test4bootsect( $lba_of_backup );
  if ($fbuf1 eq $fbuf2) {
     printf "Good!  Main and backup boot sectors are identical.  Score = %4.1f.\n", $score1;
  } else {
     # here we lower the scores if the backup boot sector is bad
     if ($score1 >= 1) {
        $score1 -= 0.1;
     }
     if ($score2 >= 1) {
        $score2 -= 0.1;
     }
     if ($score2 < 7) {
        if ($score1 >= 1) {
           $score1 -= 0.5;
        }
     }
     if (6 != $b_sectorOfBackupBootSect1) {
        print "WARNING: SectorOfBackupBootSect (at 32h) should be 6, but is actually $b_sectorOfBackupBootSect1.\n";
     }
     print "WARNING: backup boot sector at LBA $lba_of_backup does not match main boot sector at LBA $lba_offset.\n";
     printf "Main boot sector score = %4.1f, backup boot sector score = %4.1f.\n", $score1, $score2;
     if ($score2 > $score1) {
        if (($score1 < 8.5) && ($score2 > 9)) {
           print "ERROR: main boot sector is damaged, but backup boot sector at LBA $lba_of_backup is good.\n";
        } else {
           if ($score2 < 7) {
              print "ERROR: both main and backup boot sectors are in bad shape.\n";
           } else {
              print "WARNING: backup boot sector appears to be in better shape than main boot sector.\n";
           }
        }
     } else {
        if ($score1 < 7) {
           print "ERROR: both main and backup boot sectors are in bad shape";
           if ($score1 > $score2) {
              print " (but main is better than backup)";
           }
           print ".\n";
        } else {
           if ($score1 < 8.5) {
              print "Note: main boot sector might be okay";
           } else {
              print "Note: main boot sector is probably okay";
           }
           if ($score1 > $score2) {
              print " (and it's better than the backup).\n";
           } else {
              print " (and it's as good as the backup).\n";
           }
        }
     }
  }
} # sub check_bootsect_against_backup



# Note: there are four different "bases" used by Microsoft for sector numbers.
# 1) The "LBA" (Logical Block Address) is the absolute (0-based) sector
#    number on the disk.
# 2) The "relative" sector number is the sector number relative to the partition's
#    boot sector (which is relative sector no. 0).
# 3) The most obscure base for numbering sectors is the sector number relative
#    to the partition table sector which points to this partition.  The offset
#    of the partition's boot sector relative to this base is $b_numHiddenSectors.
# 4) The implied base for the data area of the partition, which would be the
#    LBA sector number of the first data sector in cluster no. 0, if there were
#    a cluster no. 0.  (Actually, the first cluster is cluster no. 2.)


sub lba_to_rel {
  local( $lba ) = @_;
  return $lba - $lba_of_bootsect;
}


sub rel_to_lba {
  local( $rel ) = @_;
  return $rel + $lba_of_bootsect;
}


sub cluster_to_rel {
  local( $clust ) = @_;
  local( $result ) = $rel_sect_of_cluster0 + ($clust * $b_sectorsPerCluster);
  return $result;
}


sub rel_to_cluster {
  local( $rel ) = @_;
  local( $result );
  $result = int( ($rel - $rel_sect_of_cluster0) / $b_sectorsPerCluster );
  return $result;
}


sub cluster_to_lba {
  local( $clust ) = @_;
  return &rel_to_lba( &cluster_to_rel( $clust ) );
}


sub lba_to_cluster {
  local( $lba ) = @_;
  return &rel_to_cluster( &lba_to_rel( $lba ) );
}


# Input is a cluster number; returns 1 if it is valid/reasonable, 0 if not.
sub is_valid_clusterno {
  local( $clust ) = @_;
  local( $result ) = 1;
  if ($clust < 2) {
     $result = 0;
  } elsif ($clust > $highest_cluster_num) {
     $result = 0;
  }
  return $result;
}


# simple byte-by-byte hex+ASCII dump of a some multiple of 16 bytes (e.g., a 512 byte sector)
sub dump_hex_and_ASCII {
  local( $sect ) = @_;
  local( $i, $j, $b, $c );
  local( $stop ) = ((!defined $sect) || ($sect eq ''));
  for ($i=0; (($i < length($sect)) && !$stop); $i+=16) {
     for ($j=0; $j < 16; $j++) {
        if (($i+$j) < length($sect)) {
           $c = substr( $sect, $i+$j, 1 );
           $b = ord($c);
           printf "%02X ", $b;
        } else {
           print "   ";
           $stop = 1;
        }
     }
     print " ";
     for ($j=0; $j < 16; $j++) {
        if (($i+$j) < length($sect)) {
           $c = substr( $sect, $i+$j, 1 );
           $b = ord($c);
           if (($b >= 32) && ($b <= 127)) {
              printf "%s", $c;
           } else {
              print ".";
           }
        }
     }
     print "\n";
  }
}


sub dump_by_sector {
   local( $cluster ) = @_;
   do {
      $sect = substr( $cluster, 0, 512 );
      $cluster = substr( $cluster, 512 );
      &dump_hex_and_ASCII( $sect );
   } until ('' eq $cluster );
}


# simple byte-by-byte hex dump of a sector, in 4-byte groups
sub dump_sect1 {
  local( $sect ) = @_;
  local( $i, $col, $b );
  for ($i=$col=0; $i < 512; $i++) {
     $b = ord(substr( $sect, $i, 1 ));
     if ($col > 0) {
        if (0 == ($i % 4)) {
           print "  ";
        } else {
           print ".";
        }
     }
     printf "%02X", $b;
     if (++$col == 32) {
        print "\n";
        $col=0;
     }
  }
}


# dump a sector as 32-bit words (good for display of FATs)
sub dump_sect2 {
  local( $sect ) = @_;
  local( $i, $col, $b, $i2 );
  for ($i=$col=0; $i < 512; $i++) {
     $i2 = $i + 3 - (2 * ($col % 4));
     $b = ord(substr( $sect, $i2, 1 ));
     if (($col > 0) && (0 == ($col % 4))) {
        print " ";
     }
     printf "%02X", $b;
     if (++$col == 32) {
        print "\n";
        $col=0;
     }
  }
}


# Read an entire FAT at once - might be many megabytes!
# Can also be used to read other parts of the disk.
# Input is the file offset in bytes, and the number of sectors to read.
# Output is a string containing everything that was read.
sub bigread1 {
  local( $fileofs, $numsect ) = @_;
  local( $result );
  local( $amt ) = (512 * $numsect);
  if ($o_debugmode) {
     printf "dbg, bigread1: seek(,$fileofs=0x0%X,0); sysread(,,$amt)\n", $fileofs;
  }
  if (!seek( DISKIMAGE, $fileofs, 0 )) {
     print "ERROR: [1] Seek(,$fileofs,0) failed, $!\n";
     exit 2;
  }
  local($rc) = sysread(DISKIMAGE, $result, $amt);
  if ($amt != $rc) {
     print "ERROR: Read failed, $amt bytes at ofs=$fileofs, rc=$rc, $!\n";
     exit 2;
  }
  if ($amt != length($result)) {
     printf "ERROR: Attempt to read $amt bytes yielded %d bytes, instead.\n", length($result);
     exit 2;
  }
  return $result;
}


# Read main FAT into $fat, but only read it once.  (To force re-read, set $fat=''.)
sub ensure_fat_loaded {
  if ((!defined $fat) || ('' eq $fat)) {
     &ensure_bootsect_loaded;  # make sure $rel_sect_of_FAT1 is set correctly
     local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
     $fat = &bigread1( $fileofs1, $b_sectorsPerFAT );
  }
}


sub compare_two_FATs {
  local( $lba_offset ) = @_;
  local( $numeq, $numne, $fsecno, $fofs, $fsec1, $fsec2 );
  $numeq = $numne = 0;
  &ensure_bootsect_loaded( $lba_offset );
  &ensure_fat_loaded;  # load first FAT into $fat
  if ($o_minmem) {
     # don't load entire 2nd FAT into memory
     for ($fsecno=0; $fsecno<$b_sectorsPerFAT; $fsecno++) {
        $fofs = 512 * $fsecno;
        $fsec1 = substr( $fat, $fofs, 512 );
        # $fsec1 = &read_lba_sect( &rel_to_lba($rel_sect_of_FAT1+$fsecno) );  # temporary
        $fsec2 = &read_lba_sect( &rel_to_lba($rel_sect_of_FAT2+$fsecno) );
        if ($fsec1 eq $fsec2) {
           $numeq++;
           if (($o_debugmode || $o_verbose) && ($fsecno < 100)) {
              # just dump the first 100 FAT sectors
              print "\nFAT sector no. $fsecno is identical in both FATs:\n";
              &dump_sect2( $fsec1 );
           }
        } else {
           $numne++;
           if (($o_debugmode || $o_verbose) && (($numne <= 20) || ($fsecno < 100))) {
              # and dump first 20 non-matching FAT sectors
              print "\nFAT sector no. $fsecno doesn't match in the two FATs.\n";
              printf "FAT #1 (LBA %d):\n", &rel_to_lba($rel_sect_of_FAT1+$fsecno);
              &dump_sect2( $fsec1 );
              printf "FAT #2 (LBA %d):\n", &rel_to_lba($rel_sect_of_FAT2+$fsecno);
              &dump_sect2( $fsec2 );
           }
        }
     }#for
     if (0 == $numne) {
        print "Good!  The two FATs are $b_sectorsPerFAT sectors each, and identical.\n";
     } else {
        print "WARNING:  The two FATs are $b_sectorsPerFAT sectors each, but NOT identical.\n";
        print "$numeq of $b_sectorsPerFAT sectors match, $numne don't.\n";
        if ($numne > 20) {
           printf "(%d mismatches not dumped)\n", $numne-20;
        }
     }
  } else {
     # normal mode -- load both FATs into memory, all at once
     # local( $fileofs1 ) = 512 * &rel_to_lba($rel_sect_of_FAT1);
     local( $fileofs2 ) = 512 * &rel_to_lba($rel_sect_of_FAT2);
     local( $fat2 ) = &bigread1( $fileofs2, $b_sectorsPerFAT );  # load 2nd FAT into $fat2
     local( $unequal ) = 0;
     if ($fat eq $fat2) {
        print "Good!  The two FATs are $b_sectorsPerFAT sectors each, and identical.\n";
     } else {
        print "WARNING:  The two FATs are $b_sectorsPerFAT sectors each, but NOT identical.\n";
        $unequal = 1;
     }
     if ($unequal || $o_debugmode || $o_verbose) {
        # The two FATs are not identical.  So compare 'em sector-by-sector:
        for ($fsecno=0; $fsecno<$b_sectorsPerFAT; $fsecno++) {
           $fofs = 512 * $fsecno;
           $fsec1 = substr( $fat, $fofs, 512 );
           $fsec2 = substr( $fat2, $fofs, 512 );
           if ($fsec1 eq $fsec2) {
              $numeq++;
              if (($o_debugmode || $o_verbose) && ($fsecno < $o_showFAT)) {
                 # just dump the first 100 FAT sectors for '-f', or 200 for '-ff', etc.
                 print "\nFAT sector no. $fsecno is identical in both FATs:\n";
                 &dump_sect2( $fsec1 );
              }
           } else {
              $numne++;
              if (($o_debugmode || $o_verbose) && (($numne <= 20) || ($fsecno < $o_showFAT))) {
                 # and dump first 20 non-matching FAT sectors
                 print "\nFAT sector no. $fsecno doesn't match in the two FATs.\n";
                 printf "FAT #1 (LBA %d):\n", &rel_to_lba($rel_sect_of_FAT1+$fsecno);
                 &dump_sect2( $fsec1 );
                 printf "FAT #2 (LBA %d):\n", &rel_to_lba($rel_sect_of_FAT2+$fsecno);
                 &dump_sect2( $fsec2 );
              }
           }
        }#for
        print "$numeq of $b_sectorsPerFAT sectors match, $numne don't.\n";
        if ($numne > 20) {
           printf "(%d mismatches not dumped)\n", $numne-20;
        }
     }
  }
} # sub compare_two_FATs


# Input is a boot sector.  Examine it to see how much it looks like a real boot
# sector, display its contents, and set a slew of global side-effects.
sub check_and_display_bootsect {
  local( $lba_offset, $score ) = @_;
  if ($o_debugmode || $o_verbose || $o_showbootsect) {
     print "============================================================\n";
  }
  &ensure_bootsect_loaded( $lba_offset );
  if ($o_debugmode) {
     &dump_hex_and_ASCII( $bootsect );
  }
  if ($o_debugmode || $o_verbose || $o_showbootsect) {
     printf "\nBOOTSECTOR-ness score = %4.1f\n", $score;
     &display_bootsect( $lba_offset );
  }
  &check_bootsect_against_backup( $lba_offset );
  if ($rel_sect_of_FAT1 == $rel_sect_of_FAT2) {
     print "Note: there is no 2nd FAT on this file system (b_numFATs=$b_numFATs).\n";
  } else {
     &compare_two_FATs( $lba_offset );
  }
} # sub check_and_display_bootsect


# Input is the LBA sector number of a boot sector.  Read the boot sector,
# set a slew of global side-effects, examine it to see how much it looks
# like a real boot sector, and display its contents.
sub read_check_and_display_bootsect {
   local( $lba_offset ) = @_;
   &ensure_bootsect_loaded( $lba_offset );  # sets $lba_of_bootsect as global side-effect
   local($score) = &test4bootsect( $lba_offset );
   &check_and_display_bootsect( $lba_of_bootsect, $score );  # sets lottsa global side-effects
} # sub read_check_and_display_bootsect


# Input is a cluster number.  This function reads the FAT entry for that
# cluster number, and does almost no checking for validity of input (so,
# for example, you can use it to fetch the dirty bits).
sub UncheckedFatEntry {
  local( $clus ) = @_;
  local( $result ) = -1;
  &ensure_fat_loaded;  # load first FAT into $fat
  if ((defined $fat) && (length($fat) > (4*$clus))) {
     $result = substr( $fat, 4 * $clus, 4 );
     ($result) = unpack( "L", $result );
  }
  return $result;
}


# Input is a cluster number.  This function reads the FAT entry for that
# cluster number, which is:
#   0          - means cluster is unallocated
#   nnn        - next cluster number in chain (if nnn >= 2)
#   0xFFFFFFF_ - cluster is last in chain
#   -1         - input was not a valid cluster number
sub FatEntry {
  local( $clus ) = @_;
  local( $result ) = -1;
  if (&is_valid_clusterno($clus)) {
     $result = &UncheckedFatEntry( $clus );
  }
  return $result;
}


# input is a cluster number
# result is an 8-character string
sub DispFatEnt {
  local( $clus ) = @_;
  local( $result );
  if ($clus >= 0) {
     $result = sprintf( "%08X", $clus );
  } else {
     $result = sprintf( "%8d", $clus );  # probably '      -1'
  }
  return $result;
}


# Given a cluster number, return the cluster number of the last cluster in
# its cluster list.  This cluster number is how we'll look up a file from any
# of its sector numbers:
# first, compute the containing cluster number from the sector number;
# then find the last cluster number for the file (using this subroutine);
# then look up the file number from its last cluster number;
# then look up the file name, directory, etc. from its file number.
sub LastClust {
  local( $nxt ) = @_;
  local( $clus ) = 0;
  $hops2last = -1;  # side-effect result
  do {
     $clus = $nxt;
     $hops2last++;
     $nxt = &FatEntry( $clus );
  } until (($nxt <= 0) || ($nxt >= 0x0FFFFFF8));
  return $clus;
}


# open the drive or image file:
if (!open( DISKIMAGE, "<$drive" )) {
   print "ERROR: Could not open '$drive', $!\n";
   exit 2;
}
binmode DISKIMAGE;


# -h option searches for likely boot sectors
if ($o_search4bootsect) {
   if (!defined $lba_offset2) {
      print "ERR: -h w/o undefined 2nd sector number\n";  # should be impossible
   } else {
      # user specified a range of sector numbers
      for ($lba_ofs=$lba_offset1; $lba_ofs <= $lba_offset2; $lba_ofs++) {
         $score = &test4bootsect( $lba_ofs );
         if ($score > 6) {
            print "\nPossible boot sector found at LBA $lba_ofs:\n";
            &read_check_and_display_bootsect( $lba_ofs );
         }
      }
   }
   exit 0;
}


# user specified just one sector number
&read_check_and_display_bootsect( $lba_offset1 );


# debug
# for ($i=600; $i<512; $i++) {
#    $tmp = &FatEntry($i);
#    printf "%03Xh: %s -> %s", $i * 4, &DispFatEnt($i), &DispFatEnt($tmp);
#    if ($tmp > 0) {
#       $tmp = &LastClust($i);
#       printf " ...[%d hops]... -> %s", $hops2last, &DispFatEnt($tmp);
#       $tmp = &FatEntry($tmp);
#       printf " -> %s", &DispFatEnt($tmp);
#    }
#    print "\n";
# }


# Is the FAT32 filesystem "dirty"?
# The Dirty indicator is stored in the high 2 bits of the 2nd FAT entry:

$dirtyness = &UncheckedFatEntry(1);
if (($dirtyness & 0x08000000) != 0) {
   print "Volume is clean.\n";
} else {
   print "*** Volume is dirty ***\n";
}
if (($dirtyness & 0x04000000) != 0) {
   print "No hard errors were seen since last mount.\n";
} else {
   print "*** HARD ERRORS HAVE BEEN SEEN ON THIS VOLUME RECENTLY! ***\n";
}


# Our inode table is represented as a set of arrays (since Perl doesn't
# have good support for arrays of structures).  We have several arrays of
# things looked up by inode (file number).  Inode #0 is the root directory,
# so we initialize the table with juts that single entry:
$in_Shortname[0] = '';  # short filename, from the directory entry
$in_ParentDir[0] = 0;  # inode of the directory it is in
$in_FirstClust[0] = $b_clusterOfRootDirStart;  # starting cluster number, from the directory entry
$in_FType[0] = 'd';  # 'f' for plain file, 'd' for directory, 'v' for volume label
# in_FSize[] has file size in bytes, usually from the directory entry (except in this case):
&LastClust( $b_clusterOfRootDirStart );
if (!$o_minmem) {
   $in_FSize[0] = (1 + $hops2last) * $b_sectorsPerCluster * 512;  # $hops2last was set by &LastClust()
}

# When reading the directory info, we scan the table from the beginning,
# searching for additional subdirectories to scan.  When we find one, we
# call &Read1dir() to add its entries to the inode table.
$nxt_inode_to_examine = 0;


# Perl pack/unpack format representing the structure of a regular 32-byte directory entry:
$DirEntry_fmt =
  'a11' . # 000h 11 bytes  Short filename (8.3, but the '.' is implied, not stored)
  'C'   . # 00Bh  1 byte   Attributes: 1=R, 2=H, 4=S, 8=volumeID, 10h=D, 20h=A, 1Fh=longName
  'C'   . # 00Ch  1 byte   reserved
  'C'   . # 00Dh  1 byte   Create time 10ths, 0-199.
  'S'   . # 00Eh  2 bytes  Create time (units = 2 sec.)
  'S'   . # 010h  2 bytes  Create date (DOS fmt)
  'S'   . # 012h  2 bytes  Last access date (DOS fmt)
  'S'   . # 014h  2 bytes  High 16 bits of 32-bit first cluster no.
  'S'   . # 016h  2 bytes  Last write time (units = 2 sec.)
  'S'   . # 018h  2 bytes  Last write date (DOS fmt)
  'S'   . # 01Ah  2 bytes  Low 16 bits of 32-bit first cluster no.
  'L';    # 01Ch  4 bytes  File size, in bytes

# Perl pack/unpack format representing the structure of a long-name directory entry:
$DirLnameEntry_fmt =
  'C'   . # 000h  1 byte   Fragment number
  'S5'  . # 001h 10 bytes  Five unicode characters of the long file name
  'C'   . # 00Bh  1 byte   Attributes: must be 1Fh (=longName)
  'C'   . # 00Ch  1 byte   reserved (must be zero)
  'C'   . # 00Dh  1 byte   Checksum of corresponding short file name
  'S6'  . # 00Eh 12 bytes  Six more unicode characters of the long file name
  'S'   . # 01Ah  2 bytes  ClustLo must be zero (corresponds to low 16 bits of cluster no. in regular dir entry)
  'S2';   # 01Ch  4 bytes  Two more unicode characters of the long file name


# rotate a byte to the right
sub RotByteRight {
   local( $i ) = @_;
   local( $result ) = 0xFF & ($i >> 1);
   if ($i & 1) {
      $result |= 0x80;
   }
   return $result;
}


# Compute checksum of 11-character (8+3) short file name, as represented in
# the short directory entry.  The "." is implied but is not part of the
# input string and not part of the calculation, however the padding blanks
# ARE part of the calculation.  (This checksum is calculated from the short
# file name, and stored in each of the corresponding long file name entries,
# as a way of validating that the long file name entries really do go with
# that short file name.)
sub CkSum {
   local( $sfname ) = @_;
   local( $b, $i, @sfnam, $result );
   $result = 0;
   @sfnam = split( //, $sfname );
   # print "dbg0: split('$sfname') = " . ($#sfnam+1) . " parts\n";
   for ($i = 0; $i < 11; $i++) {
      $b = ord( $sfnam[$i] );
      $result = (0xFF & (&RotByteRight( $result ) + $b));
   }
#   printf "dbg: sfname=\"%s\", cksum=%02X\n", $sfname, $result;
   return $result;
}


# return a 32-byte directory entry from specified offset into first input parameter
sub selDirEnt {
  local( $dirent, $ofs ) = @_;;
  if (!defined $dirent) {
     print "ERR: dirent undefined.\n";
  }
  if (!defined $ofs) {
     $ofs = 0;
  }
  $dirent = substr( $dirent, $ofs, 32 );
  if (!defined $dirent) {
     print "ERR: dirent undefined [2].\n";
  }
  return $dirent;
}


sub unpackDirEnt {
  local( $dirent ) = @_;
  local( @results );
  if (!defined $dirent) {
     print "ERR: dirent undefined.\n";
  }
  @results = unpack( $DirEntry_fmt, $dirent );
  return @results;
}


# $x = "\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F\x20";
# &unpackDirEnt_lname($x);
# exit 0;

# side-effect results are:  $ln_cksum, $ln_fragno, $ln_frag_looks_ok, and $ln_lname_piece
sub unpackDirEnt_lname {
  local( $dirent ) = @_;
  local( @tmp );
#  print "unpackDirEnt_lname:\n";
  if (!defined $dirent) {
     print "ERR: dirent undefined.\n";
  }
#  &dump_hex_and_ASCII( $dirent );
  @tmp = unpack( $DirLnameEntry_fmt, $dirent );
  if ($#tmp != 17) {
     die "ERR: unpackDirEnt_lname(), " . ($#tmp+1) . " parts instead of 18\n";
  }
#  print "dbg3: ";
#  for (local($i)=0; $i <= $#tmp; $i++) {
#     printf " %04X", $tmp[$i];
#  }
#  print "\n";
  local($ln_clustLo) = splice( @tmp, 15, 1 );
  $ln_cksum = splice( @tmp, 8, 1 );
  local($ln_resv) = splice( @tmp, 7, 1 );
  local($ln_attr) = splice( @tmp, 6, 1 );
  $ln_fragno = splice( @tmp, 0, 1 );
  $ln_frag_looks_ok = 1;
  if ((($d_attr & 0x3F) != 0xF) && ($d_attr != 0)) {
     printf "ERR: attr=0x%02X (should be 0x0F)\n", $ln_attr;
  }
  if ($ln_resv != 0) {
     if ($o_debugmode) {
        print "dbg-warn: ln_resv=$ln_resv (should be 0)\n";
     }
     $ln_frag_looks_ok = 0;
  }
  if ($ln_clustLo != 0) {
     if ($o_debugmode) {
        print "dbg-warn: ln_clustLo=$ln_clustLo (should be 0)\n";
     }
     $ln_frag_looks_ok = 0;
  }
  if ($#tmp != 12) {
     print "ERR: splice didn't remove?  " . $#tmp . " should be 12.\n";
  }
#  print "dbg4.1: ";
#  for (local($i)=0; $i <= $#tmp; $i++) {
#     printf " %04X", $tmp[$i];
#  }
#  print "\n";
  # Delete Unicode characters which do not map to legal ASCII (and also delete trailing 0-byte):
  for (local($i) = $#tmp; $i >= 0; $i--) {
     if (($tmp[$i] < 32) || ($tmp[$i] > 126)) {
        splice( @tmp, $i, 1 );
     }
  }
  $ln_lname_piece = pack( "c*", @tmp );
#  print "dbg4.2: ";
#  for (local($i)=0; $i <= $#tmp; $i++) {
#     printf " %04X", $tmp[$i];
#  }
#  print "\n";
#  &dump_hex_and_ASCII( $ln_lname_piece );
}


# Process a directory entry (but don't add it to the inode tables).
# Global side-effect results are:
#   $d_sfname = short filename
#   $d_sfname_raw = 11-byte short file name as it appears in the directory
#   $d_fsize = file size in bytes
#   $d_clust = first cluster number
#   $d_attr = attributes
#   $d_ftype = based on attributes: 'f', 'd', or other
#   $d_deleted = 1 if entry represents a deleted file, 0 otherwise
sub process1dirent {
  local( $dirent, $ofs ) = @_;
  $dirent = &selDirEnt( $dirent, $ofs );
  local( $d_resv1, $d_ctimeLo, $d_ctimeHi, $d_cdate, $d_adate, $d_clustHi, $d_wtime, $d_wdate, $d_clustLo );
  ( $d_sfname_raw, $d_attr, $d_resv1, $d_ctimeLo, $d_ctimeHi, $d_cdate, $d_adate,
    $d_clustHi, $d_wtime, $d_wdate, $d_clustLo, $d_fsize ) = &unpackDirEnt( $dirent );
  # if (!defined $d_fsize) {
  #    $d_ftype = '?';
  #    return;
  # }
  $d_clust = $d_clustLo + ($d_clustHi << 16);
  $d_sfname = $d_sfname_raw;
  $d_deleted = 0;
  local( $firstch ) = substr($dirent, 0, 1);
  if (($d_attr & 0x3F) == 0xF) {
     # it's part of a long file name
     $d_ftype = 'l';
     &unpackDirEnt_lname( $dirent );
  } elsif (   (0==$d_attr)
           && (0==$d_clustLo)
           && (($firstch eq "\xE5") || ($firstch eq "\0"))
           && (   ("\0" eq substr($dirent,2,1))
               || ("\0" eq substr($dirent,4,1))
               || ("\0" eq substr($dirent,6,1))
               || ("\0" eq substr($dirent,8,1))
               || ("\0" eq substr($dirent,10,1)) )
           && (   ("\0" ne substr($dirent,1,1))
               || ("\0" ne substr($dirent,3,1))
               || ("\0" ne substr($dirent,5,1))
               || ("\0" ne substr($dirent,7,1))
               || ("\0" ne substr($dirent,9,1)) ) ) {
     # it's part of a deleted long file name
     $d_deleted = 1;
     $d_ftype = 'l';
     &unpackDirEnt_lname( $dirent );
  } elsif (($d_attr & 8) == 8) {
     # volume label... ignore it for now
     $d_ftype = 'v';
  } elsif (0 == $d_clust) {
     # zero-length file... ignore it
     $d_ftype = 'z';
  } else {
     if (($d_attr & 0x10) == 0x10) {
        $d_ftype = 'd';  # subdirectory
     } else {
        $d_ftype = 'f';  # plain file
     }
     local( $base, $ext ) = unpack( "a8a3", $d_sfname_raw );
     if (($firstch eq "\xE5") || ($firstch eq "\0")) {
        # deleted entry
        $d_deleted = 1;
        if ($o_listdeletedfiles) {
           $base = '?' . substr($base, 1);
        } else {
           $d_ftype = '-';
           $base = $ext = $d_sfname = '';
        }
     }
     if ($d_ftype ne '-') {
        $base =~ s/\s+$//;
        if ($base =~ /^\x05/) {
           # special hack for Japanese Kanji language
           $base = "\xE5" . substr($base,1);
        }
        $ext =~ s/\s+$//;
        # print "dbg: base='$base', ext='$ext'\n";
        if ('' ne $ext) {
           $d_sfname = $base . '.' . $ext;
        } else {
           $d_sfname = $base;
        }
     }
  }
}#process1dirent


# add a directory entry to the inode tables
sub add1dirent {
  local( $parent_inode, $d_sfname, $d_fsize, $d_clust, $d_ftype ) = @_;
  local( $inode ) = $#in_Shortname + 1;
  $in_Shortname[$inode] = $d_sfname;  # short name
  $in_ParentDir[$inode] = $parent_inode;  # inode of the directory it is in
  $in_FirstClust[$inode] = $d_clust;  # starting cluster number
  $in_FType[$inode] = $d_ftype;  # 'f' for plain file, 'd' for directory
  if (!$o_minmem) {
     $in_FSize[$inode] = $d_fsize;
  }
}


sub inode2path {
  local( $inode ) = @_;
  local( $parent_inode ) = $in_ParentDir[$inode];
  local( $i, $k, $result);
  $result = '';
  $k = 0;  # protection against infinite loop
  if ($inode > $#in_Shortname) {
     $result = "***ERROR***";
  } else {
     if ($inode > 0) {
        $result = $in_Shortname[$inode];
        while ($k++ < 100) {
           $inode = $in_ParentDir[$inode];
           if (0 == $inode) {
              last;
           }
           $result = $in_Shortname[$inode] . "\\" . $result;
        }#while
     }
     $result = "\\" . $result;
  }
  return $result;
}


# When stepping through a directory, we "accumulate" the pieces of a long file
# name, until we encounter something other than the next piece of the file name
# or the corresponding short file name.  If that happens, we discard whatever
# long file name information we've thus far accumulated, by calling this
# subroutine.
sub reset_accumulating_lname {
  $accumulating_lname = '';
  $accumulating_cntr = 0;
  $prev_ftype = ' ';
  $prev_deleted = -1;
  $prev_cksum = -1;
  $prev_fragno = -1;
}


# This subroutine processes a directory (folder).
sub Read1dir {
  local( $clust1, $parent_inode ) = @_;
  local( $lastcluster ) = &LastClust( $clust1 );
  local( $DirSizeInClusters ) = 1 + $hops2last;  # $hops2last was set by &LastClust()
  # local( $DirSizeInSectors ) = $b_sectorsPerCluster * $DirSizeInClusters;
  local( $clustno ) = $clust1;
  local( $lba1, $cksum, $printthis1, $deleted_mark );
  local( $list_lnames ) = (($o_rawlistfiles >= 2) || ($o_listdeletedfiles >= 3) || $o_verbose || $o_debugmode);
  if ($o_debugmode || $o_verbose) {
     print "clust1 = $clust1 -> " . &DispFatEnt( &FatEntry($clust1) );
     print ", lastcluster = $lastcluster,  size = $DirSizeInClusters clusters ($b_sectorsPerCluster sectors/cluster)\n";
  }

  # find the full path of the parent directory
  local( $parentdir ) = &inode2path( $parent_inode );
  if (length($parentdir) > 1024) {
     print "ERROR: path too long, possible recursive directory structure, skipping directory:\n " . $parentdir . "\n";
     return;
  }

  &reset_accumulating_lname;
  do {
     if ($o_debugmode || $o_verbose) {
        printf "Cluster %d=0x%08X, which is LBA $lba1=0x0%X, which is byte offset %d=0x0%X\n",
               $clust1, $clust1, $lba1, (512 * $lba1), (512 * $lba1);
     }
     $lba1 = &cluster_to_lba($clustno);
     local( $cluster ) = &bigread1( $lba1 * 512, $b_sectorsPerCluster );
     if ($o_debugmode) {
        printf "length(cluster) = %d.\n", length($cluster);
        # &dump_by_sector( $cluster );
        if ($o_debugmode >= 2) {
           &dump_hex_and_ASCII( $cluster );
        }
     }
     for ($i=0; $i < length($cluster); $i+=32) {
        &process1dirent( $cluster, $i, $parent_inode );
        if (defined $d_sfname) {
           # found a directory entry
           if ($o_rawlistfiles || $o_listdeletedfiles || $o_verbose || $o_debugmode) {
              if ($d_deleted) {
                 $deleted_mark = '*D*';
              } else {
                 $deleted_mark = '   ';
              }
              # if '-L' or '-d' was specified, then list the directory entry
              if ('l' eq $d_ftype) {
#                 $tmp = &selDirEnt( $cluster, $i );
#                 &dump_hex_and_ASCII( $tmp );
                 if ('' ne $accumulating_lname) {
                    if (    ($ln_cksum != $prev_cksum)
                         || ($prev_ftype ne $d_ftype)
                         || ($d_deleted != $prev_deleted)
                         || !$ln_frag_looks_ok) {
                        # the checksums & deleted-ness must match for all the pieces of a long name, whether deleted or not
                        # print "dbg1: ln_cksum=$ln_cksum, prev_cksum=$prev_cksum, d_ftype=$d_ftype, prev_ftype=$prev_ftype, d_deleted=$d_deleted, prev_deleted=$prev_deleted, ln_frag_looks_ok=$ln_frag_looks_ok\n";
                        &reset_accumulating_lname;
                    } elsif (!$d_deleted) {
                        if (($ln_fragno & 0x40) || ($ln_fragno != (($prev_fragno & 0x3F) - 1))) {
                           # for non-deleted long names, the first fragment number must be OR'd with 0x40, and they count down to 1
                           # printf "dbg2: ln_fragno=0x%02X, prev_fragno=0x%02X\n", $ln_fragno, $prev_fragno;
                           &reset_accumulating_lname;
                        }
                    }
                 }
                 if ($ln_frag_looks_ok) {
                    if ('' ne $accumulating_lname) {
                       $accumulating_lname = $ln_lname_piece . $accumulating_lname;
                    } elsif ((0 != ($ln_fragno & 0x40)) || $d_deleted) {
                       $accumulating_lname = $ln_lname_piece;
                    } else {
                       # printf "dbg3: ln_fragno=0x%02X, d_deleted=%d\n", $ln_fragno, $d_deleted;
                       &reset_accumulating_lname;
                    }
                 }
                 if ('' ne $accumulating_lname) {
                    $prev_cksum = $ln_cksum;
                    $prev_fragno = $ln_fragno;
                    $prev_deleted = $d_deleted;
                    $prev_ftype = $d_ftype;
                    $accumulating_cntr++;
                 }
                 if (    ($list_lnames && !$d_deleted)
                      || (    $d_deleted
                           && (    ($list_lnames && $o_listdeletedfiles)
                                || ($o_listdeletedfiles >= 2) ) ) ) {
                    printf "%s           <LNAME> (%02X)  %02X  %d  \"%s\"",
                           $deleted_mark, $ln_cksum, $ln_fragno, $ln_frag_looks_ok, $ln_lname_piece;
                    printf "%*s%s\n", 15-length($ln_lname_piece), '', $parentdir;
                    # printf "dbg: d_deleted=$d_deleted, deleted_mark='$deleted_mark'\n";
                 }
              } else {
                 $cksum = &CkSum( $d_sfname_raw );
                 if (($prev_deleted != $d_deleted) || (($cksum != $prev_cksum) && !$d_deleted)) {
                    # print "dbg8: prev_deleted=$prev_deleted, d_deleted=$d_deleted, cksum=$cksum, prev_cksum=$prev_cksum\n";
                    &reset_accumulating_lname;
                 }
                 $printthis1 = 1;
                 if (($o_rawlistfiles && !$d_deleted) || ($o_listdeletedfiles && $d_deleted)) {
                    if ('d' eq $d_ftype) {
                       print "$deleted_mark            <DIR> ";
                    } elsif ('f' eq $d_ftype) {
                       printf "%s%11d   %02X  ", $deleted_mark, $d_fsize, $d_attr;
                    } elsif ('v' eq $d_ftype) {
                       printf "%s  Vol_Label:  %02X  ", $deleted_mark, $d_attr;
                    } else {
                       $printthis1 = 0;
                    }
                    if ($printthis1) {
                       if ((!$list_lnames) && ($o_listdeletedfiles < 2)) {
                          printf "%07X  %12s   %s", $d_clust, $d_sfname, $parentdir;
                       } elsif ($d_deleted) {
                          printf "%07X  %12s          %s", $d_clust, $d_sfname, $parentdir;
                       } else {
                          printf "%07X  %12s (%02X)     %s", $d_clust, $d_sfname, $cksum, $parentdir;
                       }
                       print "\n";
                       if ('' ne $accumulating_lname) {
                          printf "%s                           \"%s\"\n", $deleted_mark, $accumulating_lname;
                       }
                    }
                 }
              }
           }
           if (('d' eq $d_ftype) || ('f' eq $d_ftype)) {
              # files and subdirectories get added to the inode list, except for '.' and '..'
              if ((! $d_deleted) && ('.' ne $d_sfname) && ('..' ne $d_sfname)) {
                 &add1dirent( $parent_inode, $d_sfname, $d_fsize, $d_clust, $d_ftype );
              }
           }
        }#if
        if ('l' ne $d_ftype) {
           # print "dbg7: d_ftype=$d_ftype != 'l'\n";
           &reset_accumulating_lname;
        }
     }#for
     $clustno = &FatEntry($clustno);  # next cluster in chain that makes up this directory
  } while (&is_valid_clusterno($clustno));
}# sub Read1dir


# &Read1dir( $b_clusterOfRootDirStart );

if ($o_listfiles) {
   # '-l'
   print
"             size  ---cluster(s)--- path\n" .
"        ---------  ---------------- ---------\n";
}

# Read the inode list, processing each entry.  For subdirectory
# entries (type 'd') we call Read1dir to read the subdirectory,
# which adds to the end of the inode list.  So by the time we've
# gotten to the end of the inode list, we've actually read all
# the directories in the volume.
for ($inode=$nxt_inode_to_examine; $inode <= $#in_Shortname; $inode++) {
   if ('d' eq $in_FType[$inode]) {
      &Read1dir( $in_FirstClust[$inode], $inode );
   }
}

# build hash for looking up inode from the last cluster number in a cluster chain
for ($inode=0; $inode <= $#in_Shortname; $inode++) {
   $clust = $in_FirstClust[$inode];
   $LastClust2inode{ &LastClust($clust) } = $inode;
}


# display full inode/file list if '-l' was specified:
$fsize = 0;
if ($o_listfiles) {
   for ($inode=$nxt_inode_to_examine; $inode <= $#in_Shortname; $inode++) {
      if (!$o_minmem) {
         $fsize = $in_FSize[$inode];
      }
      $ftype = $in_FType[$inode];
      $clust = $in_FirstClust[$inode];
      $path = &inode2path($inode);
      $lastc = &LastClust($clust);
      $inode2 = $LastClust2inode{$lastc};
      if ($inode2 != $inode) {
         print "ERR: inode=%d -> clust=%s -> lastclust=%s -> inode=%d:\n",
               $inode, &DispFatEnt($clust), &DispFatEnt($lastc), $inode2;
      }
      printf "%5d:", $inode;
      if ($o_minmem) {
         if ('d' eq $ftype) {
            print " <DIR> ";
         } else {
            # in '-m' mode we don't store the file sizes, so we can't print 'em
            print "       ";
         }
      } else {
         if ('d' eq $ftype) {
            print "   <DIR>   ";
         } else {
            printf "%11d", $fsize;
         }
      }
      if ($lastc == $clust) {
         # just 1 cluster in cluster chain for this file
         printf "  %07X         ", $clust;
      } elsif (&FatEntry($clust) == $lastc) {
         # exactly 2 clusters in cluster chain for this file
         printf "  %07X, %07X", $clust, $lastc;
      } else {
         # 3 or more clusters in cluster chain for this file
         printf "  %07X..%07X", $clust, $lastc;
      }
    #  printf " (%d..%d) ", &cluster_to_lba($clust), &cluster_to_lba($lastc)+$b_sectorsPerCluster-1;
      print " $path\n";
   }
}


# Look up each of the sector numbers specified on the command line:
if ($o_lookup_sector) {
   print "Look up file names from sectors:\n";
   for ($i=2; $i<=$#ARGV; $i++) {
      $sect = &hex_or_positive_decimal( $ARGV[$i] );
      $clust = &lba_to_cluster( $sect );
      if ($clust < 2) {
         print "sector $sect is in the system areas.\n";
      } else {
         $nxt = &FatEntry( $clust );
         if (0 == $nxt) {
            print "sector $sect is in free space.\n";
         } elsif (0x0FFFFFF7 == $nxt) {
            printf "sector $sect is in cluster %d=0x0%X, which is marked BAD in the FAT.\n", $clust, $clust;
         } elsif (-1 == $nxt) {
            print "Could not deduce what cluster $sect is in.\n";
         } else {
            $last = &LastClust( $clust );
            $inode = $LastClust2inode{$last};
            if (!defined $inode) {
               printf "sector $sect is in cluster %d=0x0%X, which has FAT entry %s and is part of a lost cluster chain ending at cluster %d=0x0%X.\n", $clust, $clust, &DispFatEnt($nxt), $last, $last;
            } else {
               printf "sector $sect is in %s\n", &inode2path($inode);
            }
         }
      }
   }
}


exit 0;

__END__
