#!/usr/bin/perl

$| = 1;

# foreach.pl   ("Run ddrescue for each of a list of log files")
#
# Use ddrsplit.pl to split a big ddrescue log file.  Then use this Perl script
# to run ddrescue for each of the log files.
#
# This program takes a quoted command as the first parameter, and the rest of
# the parameters on the line are substituted, in turn, for the "?" in the first
# parameter.  So if there are N parameters, then the command is run N-1 times.
#
# This Perl script probably duplicates the functionality of some sort of simple,
# built-in sh (or bash) command, but I don't know bash, and I do know Perl.  :-/


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 4-Nov-2005 - first version.  -DAB
# v.2 - 6-Nov-2005 - added the -f & -t options, and delayed wildcard expansion

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "5, 13-Sep-06";
# the TLIB Version Control version number and last-modified date


$verbose = $recheck_fnames = $testmode = 0;
$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}

if (($#ARGV >= 0) && ($ARGV[0] =~ /^\-[vft]/)) {
   if ($ARGV[0] =~ /v/) {
      $verbose++;
   }
   if ($ARGV[0] =~ /f/) {
      $recheck_fnames++;
   }
   if ($ARGV[0] =~ /t/) {
      $testmode++;
   }
   shift @ARGV;
}


# Get first input parameter.
if ($#ARGV < 1) {
   die "foreach.pl  (\"repeat a command for each argument or filename\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "The command to be repeated should be the (quoted) first parameter.\n" .
       "The command will be run once for each additional parameter, with the additional\n" .
       "parameter replacing every instance of '?' in the command.\n" .
       "\n" .
       "Options:\n" .
       "   -v   Verbose\n" .
       "   -f   recheck/expand Filenames\n" .
       "   -t   Test (don't actually run the commands)\n" .
       "\n" .
       "Usage example:\n" .
       "   perl -w foreach.pl -f \"ddrescue -B /dev/hdd drive.ima ?\" \"drive.log-*\"\n" .
       "\n" .
       "For more verbose output, specify '-v', like this:\n" .
       "   perl -w foreach.pl -v \"ddrescue -B /dev/hdd drive.ima ?\" drive.log-*\n" .
       "\n" .
       "if the additional arguments are file names, you may specify '-f' to make the\n" .
       "foreach.pl program check for the existance of each file name before invoking\n" .
       "the command.\n" .
       "\n" .
       "Also, for 'late' expansion of a file name wildcard list, specify -f and quote\n" .
       "the wildcard specification(s) to let foreach.pl handle the expansion, instead\n" .
       "of letting the shell do the expansion before foreach.pl starts.  That way it\n" .
       "will find files which were added after foreach.pl started, and skip files that\n" .
       "were deleted after foreach.pl started.\n" .
       "\n" .
       "To just display the commands, and not actually run them, specify '-t'.\n";
}

if ($verbose) {
   print "You are using Perl version $perlver\n";
}

$cmd = shift @ARGV;

if ($cmd !~ /\?/) {
   die "ERROR: there must be a '?' somewhere in the quoted first argument.\n";
}


$number_done = 0;

sub do1 {
   local( $cmd, $par ) = @_;
   local( $tmpcmd ) = $cmd;
   if ($recheck_fnames) {
      # if '-f' then we avoid doing the same file more than once
      if (defined $done{$par}) {
         if ($verbose) {
            print "Skipping '$par' because we already did it.\n";
         }
         return;
      } elsif (-d $par) {
         print "Skipping '$par' because it is a directory.\n";
         return;
      } elsif (!-f $par) {
         print "Skipping '$par' because the file doesn't already exist.\n";
         return;
      }
      $done{$par} = 1;
   }
   # relace each '?' in the command with $par (probably a file name)
   $tmpcmd =~ s/\?/$par/g;
   # if ($verbose || $testmode) {
      print "run: '$tmpcmd'\n";
   # }
   if (!$testmode) {
      system( $tmpcmd );
   }
   $number_done++;
}


foreach $par (@ARGV) {
   if ($recheck_fnames) {
      if ($par =~ /\*/) {
         # it's a wildcard spec
         do {
            $prev_number_done = $number_done;
            $tmp1 = `echo $par`;  # let sh expand the wildcard spec
            $tmp1 =~ s/\s*$//g;  # remove trailing newline/whitespace
            if ($verbose) {
               print "dbg: '$par' expands to '$tmp1'\n";
            }
            @tmp = split( /\s/, $tmp1 );
            foreach $fnam (@tmp) {
               # do the command for each file that matches the wildcard spec
               &do1( $cmd, $fnam );
               # Note: As a side-effect, &do1() increments global $number_done if it actually
               # runs the command (or, in '-t' test mode, if it would have run the command).
            }
            # re-expand until wildcard expansion finds no more new files:
         } while ($number_done > $prev_number_done);
      } else {
         # '-f' mode but not wildcard, so let &do1() check that the file exists
         &do1( $cmd, $par );
      }
   } else {
      # not '-f' mode, so just do it
      &do1( $cmd, $par );
   }
}

if (0 == $number_done) {
   print "None done";
   if (!$verbose) {
      print " -- try using '$0 -v ...' to figure out why.\n";
   } else {
      print ".\n";
   }
}

__END__
